c "@(#)intg2009beta.f	1.1	2009/08/06 NGS"
c --------------------------------------------------------------------
c For further technical information, questions, or comments:
c   NOAA, National Geodetic Survey, N/NGS6, Silver Spring, MD  U.S.A.
c   Attn   : Daniel R. Roman, Ph.D.
c   Phone  : 301-713-3202
c   Fax    : 301-713-4172
c   e-mail : dan.roman@noaa.gov
c ----------------------------------------------------------------------

***********************************************************************
*                                                                     *
*                  DISCLAIMER                                         *
*                                                                     *
*   THIS PROGRAM AND SUPPORTING INFORMATION IS FURNISHED BY THE       *
* GOVERNMENT OF THE UNITED STATES OF AMERICA, AND IS ACCEPTED AND     *
* USED BY THE RECIPIENT WITH THE UNDERSTANDING THAT THE UNITED STATES *
* GOVERNMENT MAKES NO WARRANTIES, EXPRESS OR IMPLIED, CONCERNING THE  *
* ACCURACY, COMPLETENESS, RELIABILITY, OR SUITABILITY OF THIS         *
* PROGRAM, OF ITS CONSTITUENT PARTS, OR OF ANY SUPPORTING DATA.       *
*                                                                     *
*   THE GOVERNMENT OF THE UNITED STATES OF AMERICA SHALL BE UNDER NO  *
* LIABILITY WHATSOEVER RESULTING FROM ANY USE OF THIS PROGRAM.  THIS  *
* PROGRAM SHOULD NOT BE RELIED UPON AS THE SOLE BASIS FOR SOLVING A   *
* PROBLEM WHOSE INCORRECT SOLUTION COULD RESULT IN INJURY TO PERSON   *
* OR PROPERTY.                                                        *
*                                                                     *
*   THIS PROGRAM IS PROPERTY OF THE GOVERNMENT OF THE UNITED STATES   *
* OF AMERICA.  THEREFORE, THE RECIPIENT FURTHER AGREES NOT TO ASSERT  *
* PROPRIETARY RIGHTS THEREIN AND NOT TO REPRESENT THIS PROGRAM TO     *
* ANYONE AS BEING OTHER THAN A GOVERNMENT PROGRAM.                    *
*                                                                     *
***********************************************************************

      program intg2009beta

***********************************************************************
* NGS Program INTG - "INTerpolate Geoid"
* SPECIAL VERSION - for 2009 BETA data sets only. 
* Interpolates the geoid height seperation
* for a user specified position and geoid model.
* Estimates within gridded data models using either
* a spline interpolation with a 6x6 window, or
* a spline interpolation with a 4x4 window, or
* bilinear interpolation
* 
* Grided data model files are binary format
* 
* Note that "*.bin" format is binary, unformatted, direct access
* and that the order of bytes in the geoid model data files 
* --- depends on which platform the file was created ---
* Platform dependant endian condition is corrected for the binary data.
***********************************************************************
      implicit none
      CHARACTER*99  SCCSID
      CHARACTER*5   PGMVER
      CHARACTER*15  PGMDAT

      character*256 dirnam
      character*256 ifnam
      character*256 ofnam
      character*256 fnam(50)  
      character*1   ofyn
      character*1   dumyn

      character*40 b40
      character*40 sname
      character*40 xlatc
      character*40 xlonc
      character*80 card
      character*80 card2
      character*80 card86

      real*8      glamn(50)
      real*8      glamx(50)
      real*8      glomn(50)
      real*8      glomx(50)
      real*8      dla(50)
      real*8      dlo(50)
      real*8      xlat
      real*8      xlon
      real*8      fact
      real*4      val

      integer*4   ios(50)   
      integer*4   lin(50)
      integer*4   nla(50)
      integer*4   nlo(50)
      integer*4   ikind(50)

      integer  ii
      integer  linn
      integer  lout
      integer  imodel
      integer  nfiles
      integer  nff
      integer  iform
      integer  ipos
      integer  ioss
      integer  iss
      integer  kk
      integer  ight 
      integer  iinput
      integer  IS_SUBR

      logical     outfil
      logical     poseast
      logical     ne,se,we,ee

c - Variables and functions for handling endian problem
      logical   reverse_bytes(50)
      common    /endian/reverse_bytes
      real*8    frev8
      real*4    frev4
      integer*4 irev4

c - Variables for statistics of the run
c   min = smallest geoid value; minlat,minlon = lat,lon at min location
c   max = largest  geoid value; maxlat,maxlon = lat,lon at max location
c   xn  = northernmost lat;   xw = westernmost lon
c   xs  = southernmost lat;   xe = easternmost lon
      integer*4 kount,keep
      real*8    min, minlat,minlon
      real*8    max, maxlat,maxlon
      real*8    xn,xs,xw,xe
      real*8    ave,std,rms

c - Variables to get around the recl=4 issue for our real*8
c - header variables
      real*4   glamnx(2),glomnx(2),dlax(2),dlox(2)
      real*8   glamny,glomny,dlay,dloy
      equivalence(glamnx(1), glamny) 
      equivalence(glomnx(1), glomny) 
      equivalence(dlax(1),   dlay)
      equivalence(dlox(1),   dloy)

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Use common statements to cut down on the
c - amount of data sent through 'call' statements
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      common /grid1/ glamn,glamx,glomn,glomx
      common /grid2/ dla,dlo,nla,nlo,ikind
      common /grid3/ ios
      common /edges/ ne,se,we,ee

      SCCSID="@(#)intg2009beta.f	1.1	2009/08/06 NGS"
      PGMVER="1.1"
      PGMDAT="2009/08/06"

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Zero out the statistics for this run
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      ave = 0.d0
      std = 0.d0
      rms = 0.d0
      keep  = 0
      kount = 0
      max = -99999999.d0
      min = +99999999.d0
      xn = -99999.d0
      xs = +99999.d0
      xw =  99999.d0
      xe = -99999.d0
      b40 = '                    '//
     *      '                    ' 


ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Write out the introductory/disclaimer screens
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      call intro(PGMVER, PGMDAT)

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Define the file numbers
c - linn = Input file of lat/lon values
c - lout = Output file of lat/lon/geoid values
c - lin  = set of geoid files (*.bin)
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      do 1 ii=1,50
        lin(ii) = 10 + ii
    1 continue
      linn = 1
      lout = 10

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Select geoid model
c - Experimental has no prompt, yet no error message
ccccccccccccccccccccccccccccccccccccccccccccccccccc
  108 write(6,101)
  101 format(/,1x,70('-'),/,
     *   ' Which geoid model do you wish to use?',//,
     *   '  6 = USGG2009BETA',/,
     *   '  7 = GEOID09BETA                  ',/,
     *   '  ',/,
     *   ' 99 = END PROGRAM',//,
     *   '  -> ',$)
      read(5,*) imodel
      if(imodel.eq.99) stop
      if(imodel.ne.6 .and. 
     *   imodel.ne.7 )   goto 108

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Which directory are the geoid files in?      
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      call getdir(imodel, dirnam)

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Create the list of files that must be opened, and open them.
c - Return with a count of how many files were opened,
c - and a flag (ios) of which files are open and which are not.
c - IS_SUBR : run as subroutine: false=0;  true=1; (c std notation)
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      IS_SUBR = 1
      call getgrd(imodel,dirnam,IS_SUBR, nfiles,fnam,lin,nff)

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Read the headers of all geoid files which
c - where opened, and store that information.
c - Compute the max lat/lon from the header information.
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      do 115 ii=1,nfiles
        if(ios(ii).eq.0)then
          read(lin(ii),rec= 1)glamnx(1)
          read(lin(ii),rec= 2)glamnx(2)
          read(lin(ii),rec= 3)glomnx(1)
          read(lin(ii),rec= 4)glomnx(2)
          read(lin(ii),rec= 5)dlax(1)
          read(lin(ii),rec= 6)dlax(2)
          read(lin(ii),rec= 7)dlox(1)
          read(lin(ii),rec= 8)dlox(2)
          read(lin(ii),rec= 9)nla(ii)
          read(lin(ii),rec=10)nlo(ii)
          read(lin(ii),rec=11)ikind(ii)

          glamn(ii) = glamny
          glomn(ii) = glomny
          dla(ii)   = dlay
          dlo(ii)   = dloy

c Convert big endian values and to little endian
          if(ikind(ii).eq.1) then
            reverse_bytes(ii)=.FALSE.
          elseif(ikind(ii).eq.irev4(1)) then
            reverse_bytes(ii)=.TRUE.
          else
            write(6,6115) fnam(ii)
 6115       format('The file ',A,' is in an unrecognized format.',/,
     *             'intg is terminating')
            stop
          endif
          if(reverse_bytes(ii)) then
            glamn(ii)= frev8( glamn(ii) )
            glomn(ii)= frev8( glomn(ii) )
            dla(ii)  = frev8( dla(ii)   )
            dlo(ii)  = frev8( dlo(ii)   )
            nla(ii)  = irev4( nla(ii)   )
            nlo(ii)  = irev4( nlo(ii)   )
            ikind(ii)= irev4( ikind(ii) )
          endif
c
          glamx(ii) = glamn(ii) + (nla(ii)-1)*dla(ii)
          glomx(ii) = glomn(ii) + (nlo(ii)-1)*dlo(ii)
        endif
  115 continue

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - How to input?
ccccccccccccccccccccccccccccccccccccccccccccccccccc
  110    write(6,103)
  103 format(/,1x,70('-'),/,
     *       ' How would you like to input the data? ',/,
     *       ' 1 = By Keyboard (with prompts)',/,
     *       ' 2 = By File     (using allowed formats)',//,
     *       '  -> ',$)
      read(5,*)iinput
      if(iinput.lt.1 .or. iinput.gt.2)goto 110

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - If using an input file, select format.
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      if(iinput.eq.2)then
  105   write(6,104)
  104   format(/,1x,70('-'),/,
     *       ' Bluebook not supported - no geoid source code',/,
     *       ' Which format will you use for input?',/,
     *       '  1 = Free Format (For Geoid) Type 1' ,/,
     *       '  2 = Free Format (For Geoid) Type 2' ,/,
c     *       '  3 = NGS Blue Book Vol 1 (*80* and *86* records)',//,
     *       '  0 = End Program',/,
     *       ' 99 = Please explain to me the formats',//,
     *       '  -> ',$)
        read(5,*)iform    
        if(iform.eq.0)stop
        if(iform.eq.3) goto 105
ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Explain the formats if it was requested
ccccccccccccccccccccccccccccccccccccccccccccccccccc
        if(iform.eq.99)then
          call expform 
          goto 110
        endif
      endif

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Select positive longitude convention.
c - DON'T ask this if the input is Blue Book
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      if(.not.(iinput.eq.2 .and. iform.eq.3))then
 1103   write(6,1102)
 1102   format(/,1x,70('-'),/,
     *         ' Which longitude convention will you use? ',/,
     *         '   1 - Positive EAST',/,
     *         '   2 - Positive WEST',//,
     *         '  -> ',$)
        read(5,*)ipos
        if(ipos.lt.1 .or. ipos.gt.2)goto 1103
        if(ipos.eq.1)poseast = .true.  
        if(ipos.eq.2)poseast = .false.
      endif

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Open the Input file if necessary
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      if(iinput.eq.2)then
  112   write(6,109)
  109   format(/,1x,70('-'),/,
     *         ' What is your input file name? ',//,
     *         '  -> ',$)
        read(5,'(a)')ifnam
        open(linn,file=ifnam,status='old',form='formatted',
     *       iostat=ioss)
        if(ioss.ne.0)then
          write(6,111)
  111     format(' *** WARNING(111): File not found -- try again')
          goto 112
        endif
      endif
     
ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Output file? (Forced to 'yes' if there
c - was an input file)
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      if(iinput.eq.1)then
        write(6,106)
  106   format(/,1x,70('-'),/,
     *         ' Write output to a file (y/n)? ',//,
     *         '  -> ',$)
        read(5,'(a)')ofyn
      else
        ofyn = 'y'
      endif

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Open Output file if necessary
ccccccccccccccccccccccccccccccccccccccccccccccccccc
 1108 outfil = .false.
      if(ofyn.eq.'Y' .or. ofyn.eq.'y')then
        outfil = .true.
        write(6,107)
  107   format(/,1x,70('-'),/,
     *         ' Output file name?',//,
     *         '  -> ',$)
        read(5,'(a)')ofnam
        open(lout,file=ofnam,status='new',form='formatted',
     *       iostat = iss)

c - File exists...
        if(iss.ne.0)then
          write(6,1107)
 1107     format(/,' *** ERROR(1107): File Exists, try again',/)
          goto 1108
        endif
      endif


ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - Begin interpolation procedure.  Input file
c - first, then keyboard
ccccccccccccccccccccccccccccccccccccccccccccccccccc

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - USING AN INPUT FILE
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      if(iinput.eq.2)then

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - INPUT FILE, FREE FORMAT (FOR GEOID) TYPE 1
ccccccccccccccccccccccccccccccccccccccccccccccccccc
        if    (iform.eq.1) then

c - Read one line
  113     format(a80)
  116     read(linn,113,end=114)card
          kount = kount + 1

c - Find the lat/lon value 
c - Longitude always returns as 0->360...whether this
c - is positive east or west is fixed a few lines down
          call ff1(card,xlat,xlon)

c - If the lat/lon values came back as -999, set the
c - geoid value to -999 and skip the interpolation
          if(xlat.eq.-999.d0 .or. xlon.eq.-999.d0)then
            val = -999.d0
            goto 130
          endif

c - Force longitude to be positive East
          if(.not.poseast)xlon = 360.d0 - xlon

c - Find which geoid file to use, based on the lat/lon input
          call which1(xlat,xlon,nfiles,kk,imodel)

c - If the point isn't in any of our grid areas, set to -999
          if(kk.eq.-1)then
            val = -999
c - Otherwise, do the interpolation
          else            
            call interg(xlat,xlon,lin,kk, val)
            keep = keep + 1
            ave = ave + val
            rms = rms + val*val
            if(xlat.gt.xn)xn = xlat
            if(xlat.lt.xs)xs = xlat
            if(xlon.gt.xe)xe = xlon
            if(xlon.lt.xw)xw = xlon
            if(val.lt.min)then
              min = val
              minlat = xlat
              minlon = xlon
            endif
            if(val.gt.max)then
              max = val
              maxlat = xlat
              maxlon = xlon
            endif
          endif

c - Finally, write out the record to screen and possibly
c - to an output file
  130     call ff1out(iinput,outfil,lout,card,poseast,xlat,xlon,val)

c - Go back and get another value
          goto 116

c - When finished, give a little report to the 
c - screen and end program.
  114     ave = ave/keep
          rms = sqrt(rms/keep)
          if(keep.gt.1)then
            fact = dble(keep)/dble(keep-1)
            std = sqrt(fact*(rms**2 - ave**2))
          else
            std = 0
          endif
          if(poseast)then
            write(6,201)kount,keep,xn,xs,xw,xe,
     *         min,minlat,minlon,max,maxlat,maxlon,
     *         ave,std,rms
          else
            write(6,201)kount,keep,xn,xs,360.d0-xw,360.d0-xe,
     *         min,minlat,360.d0 - minlon,max,maxlat,360.d0 - maxlon,
     *         ave,std,rms
          endif 
          stop
  201     format(/,1x,70('-'),/,
     *       ' FINAL REPORT: ',/,
     *       '  Number of points input: ',i8,/,
     *       '  Number of good points : ',i8,/,
     *       '  Northernmost Latitude : ',f10.6,/,
     *       '  Southernmost Latitude : ',f10.6,/,
     *       '  Westernmost Longitude : ',f10.6,/,
     *       '  Easternmost Longitude : ',f10.6,/,
     *       '  Minimum Geoid Height  : ',f8.3,/,
     *       '    Lat/Lon of Minimum  : ',f10.6,1x,f10.6,/,
     *       '  Maximum Geoid Height  : ',f8.3,/,
     *       '    Lat/Lon of Maximum  : ',f10.6,1x,f10.6,/,
     *       '  Average Geoid Height  : ',f8.3,/,
     *       '  Standard Deviation    : ',f8.3,/,
     *       '  Root Mean Square      : ',f8.3,//)

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - INPUT FILE, FREE FORMAT (FOR GEOID) TYPE 2
ccccccccccccccccccccccccccccccccccccccccccccccccccc
        elseif(iform.eq.2)then

c - Read one line
  117     format(a80)
  119     read(linn,117,end=118)card
          kount = kount + 1

c - Find the lat/lon value
c - Longitude always returns as 0->360...whether this
c - is positive east or west is fixed a few lines down
          call ff2(card,xlat,xlon)

c - If the lat/lon values came back as -999, set the
c - geoid value to -999 and skip the interpolation
          if(xlat.eq.-999.d0 .or. xlon.eq.-999.d0)then
            val = -999.d0
            goto 330
          endif

c - Force longitude to be positive East
          if(.not.poseast)xlon = 360.d0 - xlon

c - Find which geoid file to use, based on the lat/lon input
          call which1(xlat,xlon,nfiles,kk,imodel)

c - If the point isn't in any of our grid areas, set to -999
          if(kk.eq.-1)then
            val = -999
          else
c - Otherwise, do the interpolation
            call interg(xlat,xlon,lin,kk, val)
            keep = keep + 1
            ave = ave + val
            rms = rms + val*val
            if(xlat.gt.xn)xn = xlat
            if(xlat.lt.xs)xs = xlat
            if(xlon.gt.xe)xe = xlon
            if(xlon.lt.xw)xw = xlon
            if(val.lt.min)then
              min    = val
              minlat = xlat
              minlon = xlon
              endif
            if(val.gt.max)then
              max    = val
              maxlat = xlat
              maxlon = xlon
            endif
          endif

c - Finally, write out the record to screen and possibly
c - to an output file
  330       call ff2out(outfil,lout,card,poseast,xlat,xlon,val)

c - Go back and get another value
          goto 119

c - When finished, give a little report to the 
c - screen and end program.
  118     ave = ave/keep
          rms = sqrt(rms/keep)
          if(keep.gt.1)then
            fact = dble(keep)/dble(keep-1)
            std = sqrt(fact*(rms**2 - ave**2))
          else
            std = 0
          endif
          if(poseast)then
            write(6,201)kount,keep,xn,xs,xw,xe,
     *         min,minlat,minlon,max,maxlat,maxlon,
     *         ave,std,rms
          else  
            write(6,201)kount,keep,xn,xs,360.d0-xw,360.d0-xe,
     *         min,minlat,360.d0-minlon,max,maxlat,360.d0-maxlon,
     *         ave,std,rms
          endif
          stop         


ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - INPUT FILE, BLUE BOOK FORMAT
ccccccccccccccccccccccccccccccccccccccccccccccccccc
        endif

ccccccccccccccccccccccccccccccccccccccccccccccccccc
c - USING KEYBOARD INPUT - Prompts
ccccccccccccccccccccccccccccccccccccccccccccccccccc
      else

c - Prompt user for station name
  133   write(6,131)
  131   format(/,1x,70('-'),/,
     *       ' What is the name of this point?',//,
     *       '  -> ',$)
        read(5,'(a)')sname
        kount = kount + 1

c - Prompt user for latitude
  502   write(6,132)
  132   format(/,1x,70('-'),/,
     *       ' What is the latitude of this point?',//,
     *       '  -> ',$)
        read(5,'(a)')xlatc
        call c2v(xlatc,xlat,1)
        if(xlat.eq.-999.d0)then
          write(6,501)
  501     format(' *** ERROR(501): Bad Latitude...try again...')
          goto 502
        endif

c - Prompt user for longitude, using their desired convention
  503   if(poseast)     write(6,135)
        if(.not.poseast)write(6,1135)

  135   format(/,1x,70('-'),/,
     *       ' What is the longitude of this point?',/,
     *       ' (Positive East, 0->360)',//
     *       '  -> ',$)
 1135   format(/,1x,70('-'),/,
     *       ' What is the longitude of this point?',/,
     *       ' (Positive West, 0->360)',//
     *       '  -> ',$)

c - Convert the card into variable "xlon"
c - Longitude always returns as 0->360...whether this
c - is positive east or west is fixed a few lines down
        read(5,'(a)')xlonc
        call c2v(xlonc,xlon,2)
        if(xlon.eq.-999.d0)then
          write(6,504)
  504     format(' *** ERROR(504): Bad Longitude...try again...')
          goto 503
        endif

c - Force the longitude to be positive EAST
        if(.not.poseast)xlon = 360.d0 - xlon

c - Find which geoid file to use, based on the lat/lon input
            call which1(xlat,xlon,nfiles,kk,imodel)

c - If the point isn't in any of our grid areas, set to -999
            if(kk.eq.-1)then
              val = -999
            else            
c - Otherwise, do the interpolation
              call interg(xlat,xlon,lin,kk, val)
              keep = keep + 1
              ave = ave + val
              rms = rms + val*val
              if(xlat.gt.xn)xn = xlat
              if(xlat.lt.xs)xs = xlat
              if(xlon.gt.xe)xe = xlon
              if(xlon.lt.xw)xw = xlon
              if(val.lt.min)then
                min = val
                minlat = xlat
                minlon = xlon
              endif
              if(val.gt.max)then
                max = val
                maxlat = xlat
                maxlon = xlon
              endif
            endif

c - Finally, write out the record to screen and possibly
c - to an output file.  Use Free Format 1 output
            card = sname//b40
            call ff1out(iinput,outfil,lout,card,poseast,xlat,xlon,val)

c - Ask user if another point is to be used, and 
c - if so, go back and get it.
            write(6,134)
  134       format(/,1x,70('-'),/,
     *       ' Compute another?',//
     *       '  -> ',$)
        read(5,'(a)')dumyn
        if(dumyn.eq.'y' .or. dumyn.eq.'Y')goto 133

c - When finished, give a little report to the 
c - screen and end program.
  121   ave = ave/keep
        rms = sqrt(rms/keep)
        if(keep.gt.1)then
          fact = dble(keep)/dble(keep-1)
          std = sqrt(fact*(rms**2 - ave**2))
        else
          std = 0
        endif
        if(poseast)then
          write(6,201)kount,keep,xn,xs,xw,xe,
     *       min,minlat,minlon,max,maxlat,maxlon,
     *       ave,std,rms
        else
          write(6,201)kount,keep,xn,xs,360.d0-xw,360.d0-xe,
     *       min,minlat,360.d0-minlon,max,maxlat,360.d0-maxlon,
     *       ave,std,rms
        endif
        stop         
      endif


ccccccccccccccccccccccccccccccccccccccccccccccccccc      
c - End Of Main Program
ccccccccccccccccccccccccccccccccccccccccccccccccccc
 
      end
c
